###############################################################################
# Auto-generated by `jupyter-book config`
# If you wish to continue using _config.yml, make edits to that file and
# re-generate this one.
###############################################################################

import larch

html_static_path = ["_static"]

html_css_files = [
    "css/custom.css",
]

add_module_names = False
author = "The Jupyter Book community"
autodoc_default_options = {"show-inheritance": True}
autosummary_generate = True
bibtex_bibfiles = ["references.bib"]
bibtex_reference_style = "author_year"
comments_config = {"hypothesis": False, "utterances": False}
copyright = "2024"
exclude_patterns = ["**.ipynb_checkpoints", ".DS_Store", "Thumbs.db", "_build"]
extensions = [
    "sphinx_togglebutton",
    "sphinx_copybutton",
    "myst_nb",
    # "jupyter_book",
    "pydata_sphinx_theme",
    "sphinx_thebe",
    "sphinx_comments",
    # "sphinx_external_toc",
    "sphinx.ext.intersphinx",
    "sphinx_design",
    # "sphinx_book_theme",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "sphinx.ext.doctest",
    "sphinx.ext.graphviz",
    "sphinx.ext.viewcode",
    "sphinxcontrib.bibtex",
    # "sphinx_jupyterbook_latex",
]
# external_toc_exclude_missing = False
# external_toc_path = "_toc.yml"
html_baseurl = ""
html_favicon = "_img/larch_favicon.png"
# html_logo = "_img/larch-logo.png"
html_show_copyright = False
html_sourcelink_suffix = ""
html_theme = "pydata_sphinx_theme"
html_theme_options = {
    "search_bar_text": "Search these docs...",
    "favicons": [{"rel": "icon", "sizes": "32x32", "href": "_img/larch_favicon.png"}],
    "navigation_with_keys": True,
    "logo": {
        "text": f"v{larch.__version__}",
        "image_light": "_img/larch-logo.png",
        "image_dark": "_img/larch-logo.png",
    },
}
html_title = f"v{larch.__version__}"
latex_engine = "pdflatex"
myst_enable_extensions = [
    "colon_fence",
    "dollarmath",
    "linkify",
    "substitution",
    "tasklist",
]
myst_url_schemes = ["mailto", "http", "https"]
napoleon_use_rtype = False
nb_execution_allow_errors = True
nb_execution_cache_path = ""
nb_execution_excludepatterns = []
nb_execution_in_temp = False
nb_execution_mode = "cache"
nb_execution_timeout = 300
nb_output_stderr = "show"
numfig = True
pygments_style = "sphinx"
rst_epilog = (
    ".. |idca| replace:: :ref:`idca <idca>`\n.. |idco| replace:: :ref:`idco <idco>`\n"
)
suppress_warnings = ["myst.domains"]
use_jupyterbook_latex = True
use_multitoc_numbering = True
