
# Example code extracted automatically from documentation.
# Do not edit these files manually.
from ._030_mnl_constrain import example as _030_mnl_constrain
from ._001_mnl import example as _001_mnl
from ._017_mnl import example as _017_mnl
from ._022_nl import example as _022_nl
from ._151_swissmetro_panel_latent_class import example as _151_swissmetro_panel_latent_class
from ._102_swissmetro_weighted import example as _102_swissmetro_weighted
from ._109_swissmetro_nl import example as _109_swissmetro_nl
from ._101_swissmetro_mnl import example as _101_swissmetro_mnl
from ._204_exville_dest_only import example as _204_exville_dest_only
from ._251_bySpec import example as _251_bySpec
from ._200_exampville import example as _200_exampville
from ._203_exville_dest import example as _203_exville_dest
from ._201_exville_mode_choice import example as _201_exville_mode_choice
from ._202_exville_mc_logsums import example as _202_exville_mc_logsums
from ._xlogit_test import example as _xlogit_test

ex = {}
ex[30] = _030_mnl_constrain
ex[1] = _001_mnl
ex[17] = _017_mnl
ex[22] = _022_nl
ex[151] = _151_swissmetro_panel_latent_class
ex[102] = _102_swissmetro_weighted
ex[109] = _109_swissmetro_nl
ex[101] = _101_swissmetro_mnl
ex[204] = _204_exville_dest_only
ex[251] = _251_bySpec
ex[200] = _200_exampville
ex[203] = _203_exville_dest
ex[201] = _201_exville_mode_choice
ex[202] = _202_exville_mc_logsums
ex['xlogit'] = _xlogit_test
